@@ -17,6 +17,8 @@
 # under the License.
 #
 from io import BytesIO
+from queue import Queue
+from threading import Lock, Thread
 from typing import Any, List, Optional
 
 import boto3
@@ -86,12 +88,6 @@ class OpenstackHook(BaseHook):
         self.conn = None
 
     def list_volumes(self) -> List[any]:
-        """
-        Returns a list of files on the remote system.
-
-        :param path: full path to the remote directory to list
-        :type path: str
-        """
         conn = self.get_conn()
         return conn.list_volumes()
 
@@ -161,50 +157,85 @@ class OpenstackHook(BaseHook):
                 f.write(chunk)
         return file_path
 
-    def stream_image_to_s3(self, imgid: str,  aws_conn_id: str, bucket: str, part_size_mb: int):
+    def stream_image_to_s3(self,
+                           imgid: str,
+                           aws_conn_id: str,
+                           bucket: str,
+                           part_size_mb: int,
+                           threads_number: int = 2):
         conn = self.get_conn()
         img = conn.get_image_by_id(imgid)
         filename = "{}_{}.img".format(img.name, img.id)
 
         s3_hook = S3Hook(aws_conn_id=aws_conn_id)
         s3: boto3.resource = s3_hook.get_resource_type("s3")
-        s3Client: boto3.client = s3_hook.get_client_type("s3")
+        s3client: boto3.client = s3_hook.get_client_type("s3")
 
-        mpu = s3Client.create_multipart_upload(
+        mpu = s3client.create_multipart_upload(
             Bucket=bucket,
             Key=filename
         )
 
         stream = BytesIO()
+        part_lock = Lock()
+        queue = Queue(threads_number)
+
         parts = []
-        partcount = [0]
+        partnum = 1
 
         def uploadPart():
-            partcount[0] += 1
-            print("uploading part %i" % partcount[0])
-            stream.seek(0)
-            upload_part = s3.MultipartUploadPart(
-                bucket, filename, mpu['UploadId'], partcount[0]
-            )
-            upload_part_response = upload_part.upload(
-                Body=stream,
-            )
-            parts.append({
-                'PartNumber': partcount[0],
-                'ETag': upload_part_response['ETag']
-            })
-            stream.seek(0)
-            stream.truncate()
+            while True:
+                val = queue.get()
+                if val is None:
+                    queue.task_done()
+                    return
+
+                stream = val['stream']
+                partcount = val['partcount']
+                print("uploading part %i" % partcount)
+                stream.seek(0)
+                uploadPart = s3.MultipartUploadPart(
+                    bucket, filename, mpu['UploadId'], partcount
+                )
+                uploadPartResponse = uploadPart.upload(
+                    Body=stream,
+                )
+                with part_lock:
+                    parts.append({
+                        'PartNumber': partcount,
+                        'ETag': uploadPartResponse['ETag']
+                    })
+                stream.seek(0)
+                stream.truncate()
+                print("Part %i uploaded" % partcount)
+                queue.task_done()
+
+        for i in range(threads_number):
+            Thread(target=uploadPart, daemon=True).start()
 
         response = conn.image.download_image(imgid, True)
         for chunk in response.iter_content(chunk_size=part_size_mb << 20):  # until EOF
             if not chunk:  # EOF?
                 break
             stream.write(chunk)
-            if stream.tell() > part_size_mb << 20:
-                uploadPart()
-        uploadPart()
-        s3Client.complete_multipart_upload(Bucket=bucket, Key=filename,
+            if stream.tell() >= part_size_mb << 20: # We have to test this as chunk_size could be working improperly
+                print("part %i downloaded" % partnum)
+                queue.put({'partcount': partnum, 'stream': stream})
+                partnum += 1
+                stream = BytesIO()
+        queue.put({'partcount': partnum, 'stream': stream})
+
+        # We send None to say the upload is finished
+        print("waiting for upload to finish")
+        # Send none for the threads to stop
+        for i in range(threads_number):
+            queue.put(None)
+        queue.join()
+
+        # Reorder parts by number
+        parts.sort(key=lambda x: x['PartNumber'])
+
+        s3client.complete_multipart_upload(Bucket=bucket, Key=filename,
                                            MultipartUpload={'Parts': parts},
                                            UploadId=mpu['UploadId'])
 
